 /*
 * @(#)mydoomscan.c v1.0 27 JAN 2004 Rob Thomas robt@cymru.com
 * Team Cymru <http://www.cymru.com/>
 * A multithreaded scanner for hosts infected with W32.MyDoom,
 * aka Novarg.
 */

#define _REENTRANT

#include "mydoomscan.h"

typedef struct {
	char ipaddress[1024];
	FILE *output;
} threadWorkOrder_t;

threadWorkOrder_t *workOrder;
int id = 0;
int no_res = 0;
pthread_t myDoomThread[255];

void timedout(int signo);
void usage();
int myDoom(threadWorkOrder_t *);

int
main(int argc, char *argv[])
{
   extern int optopt;
   extern char *optarg;
   int c = 0;
   int range = 0;
   int list = 0;
   int singlehost =0;
   int max_thread = 1;
   int timeout = 15;
   int high = 0;
   int low = 0;
   int loop = 0;
   int herrno;
   struct sockaddr_in sin; 
   unsigned long start;
   unsigned long end;
   unsigned long counter;
   char outputfile[1024];
   char ipaddress[1024];
   char startip[1024];
   char endip[1024]; 
   char buf[8192];
   struct hostent *host;
   struct hostent result;
   struct in_addr inaddr;
   FILE *input, *output;
   struct rlimit limit;

   limit.rlim_max = 1024;
   limit.rlim_cur = 1024;
   if(setrlimit(RLIMIT_NOFILE, &limit)) {
#ifdef VERBOSE
		fprintf(stderr, "setrlimit() failed for nofiles.\n");
#endif
   }

   if((argc < 3) || (argc > 10))
      usage();
   while((c = getopt(argc,argv,"nh:s:e:l:t:c:")) != -1)
   {
      switch(c)
      {
		 case 'n' : /* No name resolution please */
		 {
			no_res = 1;
			break;
		 }
         case 'h' :  /* Single host scan  */
         {
			if(inet_pton(AF_INET, optarg, &inaddr)) {
				host = NULL;
	            if((start = inet_addr(optarg))==INADDR_NONE) {
					fprintf(stderr,"%s is an invalid IP Address.\n",optarg);
					exit(-1);
				}
			} else {
				if (no_res == 0) {
					host = gethostbyname(optarg);
				} else {
					fprintf(stderr,
					 "ERROR:  Name given but name resolution is disabled!\n\n");
					usage();
					exit(-1);
				}
			}
            sprintf(outputfile, "%s.txt", optarg);
            sprintf(startip, "%s", optarg);
            singlehost = 1;
            break;
         } 
         case 's' :  /* Starting ip address */
         {
            if((start = inet_addr(optarg))==INADDR_NONE)
            {
               fprintf(stderr,"%s is an invalid IP Address.\n",optarg);
               exit(-1);
            }
            sprintf(startip,"%s",optarg);
            break;
         }
         case 'e' : /* Ending ip address */
         {
            if((end = inet_addr(optarg))==INADDR_NONE)
            {
               fprintf(stderr,"%s is an invalid IP Address.\n",optarg);
               exit(-1);
            }
            sprintf(endip,"%s",optarg);
            sprintf(outputfile,"%s-%s.txt",startip,endip);
            range = 1;
            break;
         }
         case 'l' :
         {
            if((input = fopen(optarg, "r")) == NULL)
            {
                printf("Unable to open %s\n",optarg);
                exit(-1);
            }
            sprintf(outputfile,"results-%s",optarg); 
            list = 1;
            break;
         }
		 case 't' : /* Maximum number of concurrent threads */
		 {
			if(atoi(optarg) > 255) {
	 	 	 	fprintf(stderr, "Maximum threads = 255, setting to 255.\n");
				max_thread = 255;
			} else if(atoi(optarg) < 1) {
				max_thread = 1;
	 	 	} else {
				max_thread = atoi(optarg);
			}
	 	 	break;
		 }
		 case 'c' : /* Scan timeout */
		 {
			if(atoi(optarg) < 5) {
			   fprintf(stderr, "Minimum timeout is 5 seconds, setting to 5.\n");
			   timeout = 5;
			} else {
				timeout = atoi(optarg);
			}
			break;
		 }
         default:
         {
            usage(); 
            break;
         }
        
      }
   }

  if((output=fopen(outputfile,"a")) == NULL)
  {
     printf("Unable to open %s for writing.\n",outputfile);
     exit(-1);
  } 

  if(singlehost)
  {
	bzero(&sin, sizeof(sin));
	counter = ntohl(start);

	if (host) {
		memcpy(&sin.sin_addr, host->h_addr_list[0], sizeof(sin.sin_addr));
	} else {
		sin.sin_addr.s_addr = htonl(counter);
	}
	sprintf(ipaddress, "%s", inet_ntoa(sin.sin_addr));
	workOrder = (threadWorkOrder_t *)malloc(sizeof(threadWorkOrder_t));
	sprintf(workOrder->ipaddress, "%s", ipaddress);
	workOrder->output = output;
	pthread_create(&myDoomThread[id], NULL, (void *) myDoom, (void *) workOrder);
	signal(SIGALRM, timedout);
	alarm(timeout);
	pthread_join(myDoomThread[loop], NULL);
  }
 
  if(range) {
		 low = (-1 * ntohl(end));
		 high = (-1 * ntohl(start));
		 counter = ntohl(start);
		 /* Fixed bug here. */
		 while (counter < ntohl(end)) {
			if((high-low) < max_thread) {

				for(loop = low; loop <= high; loop++) {				
					bzero(&sin, sizeof(sin));
					sin.sin_addr.s_addr = htonl(counter);      
					sprintf(ipaddress, "%s", inet_ntoa(sin.sin_addr));
					workOrder =
						(threadWorkOrder_t *)malloc(sizeof(threadWorkOrder_t)); 
					sprintf(workOrder->ipaddress, "%s", ipaddress);
					workOrder->output = output;
					pthread_create(&myDoomThread[id], NULL,
						(void *) myDoom, (void *) workOrder);
					id++;
					counter++;
				}

        		signal(SIGALRM, timedout);
        		alarm(timeout);

				for(loop = 0; loop < id; loop++) {				
					pthread_join(myDoomThread[loop], NULL);
				}

				low = low + max_thread;
				id = 0;
			} else {
				for(loop = low; loop <= (low+max_thread); loop++)
				{
                    bzero(&sin, sizeof(sin));
                    sin.sin_addr.s_addr = htonl(counter);
                    sprintf(ipaddress,"%s", inet_ntoa(sin.sin_addr));
                    workOrder =
                        (threadWorkOrder_t *)malloc(sizeof(threadWorkOrder_t));
                    sprintf(workOrder->ipaddress, "%s", ipaddress);
                    workOrder->output = output;
                    pthread_create(&myDoomThread[id], NULL,
                        (void *) myDoom, (void *) workOrder);
					id++;
					counter++;
				}

        		signal(SIGALRM, timedout);
        		alarm(timeout);

                for(loop = 0; loop < id; loop++)
                {
					pthread_join(myDoomThread[loop], NULL);
				}
				low = low + max_thread;
				id = 0;
			}
		 }
  }

  if(list) {
     while(fgets(ipaddress,sizeof(ipaddress),input)!=NULL) {
         char *token;
         token = strtok(ipaddress, "\n");
         c = 0; 

         while (token !=NULL) {

            switch(c) {
                case 0:
                sprintf(ipaddress, "%s",token);
                break;
                default:
                break;
            }

            token =strtok(NULL,"\n");
			
			workOrder = (threadWorkOrder_t *)malloc(sizeof(threadWorkOrder_t));
			sprintf(workOrder->ipaddress, "%s", ipaddress);
			workOrder->output = output;
			pthread_create(&myDoomThread[id], NULL, (void *) myDoom, (void *) workOrder);
            c = c +1;
			id++;
			if(id >= max_thread) {

				for(loop = 0; loop < id; loop++) {
					pthread_join(myDoomThread[loop], NULL);
				}

				id = 0;
			}
         }
     }

	 if(id != 0) {

		 for(loop = 0; loop < id; loop++) {
			pthread_join(myDoomThread[loop], NULL);
		 }
	 }
  }

  fprintf(stdout,"Done.\n");
/*  fprintf(output,"Done.\n"); */
  fclose(output);
  return(0);
}

int
myDoom(threadWorkOrder_t *workOrder)
{
    struct sockaddr_in sin;
    struct hostent *host, result;
    char sendbuff[BUFFSIZE];
    char recvbuff[BUFFSIZE];
	char buf[8192];
    char *line;
    struct in_addr inaddr;
    int sockfd = 0;
    int i  = 0;    
    int infected = 1; 
	int herrno;
    char *payload[] ={"mydoomscan.c checker 1.0\n"};
    struct timeval tv;

    bzero(&sin,sizeof(sin));

	if (no_res == 0) {
#if defined(__GLIBC__)
	    struct hostent hent;
	    char hbuf[8192];

	    if(inet_pton(PF_INET, workOrder->ipaddress, &inaddr)) {
		    if (gethostbyaddr_r((char *) &inaddr,
			sizeof(inaddr), PF_INET, &hent, hbuf,
			sizeof(hbuf), &host, &herrno) < 0)
		    {
		        host = NULL;
    		    }
		} else {
		    if (gethostbyname_r(workOrder->ipaddress, &hent, hbuf,
				sizeof(hbuf), &host, &herrno) < 0)
		    {
		        host = NULL;
		    }
		}	
#elif defined(sun)
	    if(inet_pton(AF_INET, workOrder->ipaddress, &inaddr)) {
			host = gethostbyaddr_r((char *) &inaddr, sizeof(inaddr), AF_INET,
					&result, buf, sizeof(buf), &herrno);
	    } else {
			host = gethostbyname_r(workOrder->ipaddress, &result, buf,
					sizeof(buf), &herrno);
		}
#elif defined(AIX)
        struct hostent_data *hent_data;
        if(inet_pton(AF_INET, workOrder->ipaddress, &inaddr)) {
            if (gethostbyaddr_r((char *) &inaddr, sizeof(inaddr), AF_INET,
                    host, hent_data) < 0) {
                host = NULL;
            }
        } else {
            if (gethostbyname_r(workOrder->ipaddress, host, hent_data) < 0) {
                host = NULL;
            }
        }
#else
#error "I don't know the reentrant lookup calls on this system."
#endif
	} else {
		host = NULL;
	}

    if(host) {
		memcpy(&sin.sin_addr, host->h_addr_list[0], sizeof(sin.sin_addr));
    } else {
		if((inet_addr(workOrder->ipaddress))!=INADDR_NONE) {
			sin.sin_addr.s_addr = inet_addr(workOrder->ipaddress);
		} else {
			fprintf(stderr, "Eject!  Eject!  No IP address to scan!\n");
			exit(1);
		}
	}

/*    for(i=0;i<=5 && (infected==1);i++) { */
       if((sockfd = socket(AF_INET,SOCK_STREAM,0)) == -1) {
           fprintf(stderr,"Unable to initialize socket.\n");
           return;
        }

        tv.tv_sec  = SNDRCVTIMEOUT;
        tv.tv_usec = SNDRCVTIMEOUT; 
        #if defined (_BSD)
        if(setsockopt(sockfd,SOL_SOCKET,SO_RCVTIMEO,&tv,sizeof(tv))!=0)
           perror("setsockopt");
        if(setsockopt(sockfd,SOL_SOCKET,SO_SNDTIMEO,&tv,sizeof(tv))!=0)
           perror("setsockopt");
        if(setsockopt(sockfd,SOL_SOCKET,SO_RCVLOWAT,(void *) &wait,sizeof(wait))!=0)
           perror("setsockopt");
#endif
        
        sin.sin_family=AF_INET;
        sin.sin_port = htons(3127);


        if((connect(sockfd, (struct sockaddr *)&sin, sizeof(sin)) == -1)) {
           if(errno==EINTR)
              errno=ETIMEDOUT;
			#ifdef VERBOSE
			fprintf(stderr,"%s Closed\n",(char *)inet_ntoa(sin.sin_addr));
			#endif
           close(sockfd);
           return;
        }

	#ifdef VERBOSE
	fprintf(stderr,"%s Open\n",(char *)inet_ntoa(sin.sin_addr));
	#endif
        memset(recvbuff,0,BUFFSIZE);
        memset(sendbuff,0,BUFFSIZE);
        
        sprintf(sendbuff,"%s",payload[i]);
      
        if(send(sockfd,sendbuff,strlen(sendbuff),0) == -1) {
           fprintf(stderr, "Error sending request.\n");
           close(sockfd);
           return;
        }
		#ifdef VERBOSE
		else
		fprintf(stderr,"%s",sendbuff);
		#endif
    
        while(recv(sockfd,recvbuff,BUFFSIZE,0)>0) {
			#ifdef VERBOSE
			fprintf(stderr,"%s",recvbuff);
			#endif
           
           line = strtok(recvbuff,"\r\n");

                 if(strstr(line,"[")) {
					fprintf(stdout,"%s Infected with W32.MyDoom\n", (char *)inet_ntoa(sin.sin_addr));
					fprintf(workOrder->output,"%s Infected with W32.MyDoom\n", (char *)inet_ntoa(sin.sin_addr));
                 } else {
						infected = 0;
						fprintf(stdout,"%s Clean\n", (char *)inet_ntoa(sin.sin_addr));
						fprintf(workOrder->output,"%s Clean\n", (char *)inet_ntoa(sin.sin_addr));
		}
		break;
              line = strtok(NULL,"\r\n"); 
        } 

        close(sockfd);
        fflush(workOrder->output); 
		#ifdef VERBOSE
		fprintf(stderr,"Closing %s\n",(char *)inet_ntoa(sin.sin_addr));
		#endif
/*    } */
    return(0);  
}

void
timedout(int signo)
{
	int loop = 0;

	signal(SIGALRM, timedout);

	#ifdef VERBOSE
	fprintf(stderr,"Connection timed out.  Detaching %d threads.\n", id);
	#endif

	for(loop = 0; loop <= id; loop++)
	{				
		pthread_detach(myDoomThread[loop]);
		#ifdef VERBOSE
		fprintf(stderr, "pthread_detach() sent to thread %d.\n", id);
		#endif
	}
	return;
}

void
usage()
{
	fprintf(stderr,"mydoomscan.c v1.0 27 JAN 2004 Rob Thomas robt@cymru.com\n");
	fprintf(stderr,"Team Cymru <http://www.cymru.com/>\n");
	fprintf(stderr, "Usage: mydoomscan [-hseltcn]\n");
	fprintf(stderr, "\t-h <hostname|IP> - Scan a single host.\n");
	fprintf(stderr, "\t-s <start IP> -e <end IP> - Scan a range of hosts.\n");
	fprintf(stderr, "\t-l <file> - Scan a list of IPs contained in <file>.\n");
	fprintf(stderr, "\t-t <threads> - Maximum concurrent threads, no higher than 255.\n");
	fprintf(stderr, "\t-c <seconds> - Timeout in seconds, default 15.\n");
	fprintf(stderr, "\t-n - No name resolution.\n");
	exit(-1);
}

